﻿using System;
using System.Collections;
using System.Linq.Expressions;
using System.Reflection;

namespace NullConditionalOperator
{
    static class NullConditionalExtension
    {
        public static object GetInternalRecursive(object obj, Expression expression)
        {
            if (obj == null)
                return null;

            if (expression.NodeType == ExpressionType.MemberAccess)
            {
                var member = (MemberExpression)expression;
                obj = GetInternalRecursive(obj, member.Expression);
                if (obj != null)
                {
                    if (member.Member is PropertyInfo)
                        obj = ((PropertyInfo)member.Member).GetValue(obj);
                    else
                        obj = ((FieldInfo)member.Member).GetValue(obj);
                }
                return obj;
            }
            else if (expression.NodeType == ExpressionType.Constant)
            {
                return ((ConstantExpression)expression).Value;
            }
            else if (expression.NodeType == ExpressionType.ArrayIndex)
            {
                var binary = (BinaryExpression)expression;
                var left = GetInternalRecursive(obj, binary.Left);
                var right = GetInternalRecursive(obj, binary.Right);
                if (left != null && right != null)
                    return ((IList)left)[(int)right];
                return
                    null;
            }
            else if (expression.NodeType == ExpressionType.Parameter)
            {
                return obj;
            }
            else
            {
                throw new ArgumentException();
            }
        }


        public static int? Get<T>(this T obj, Expression<Func<T, int>> expr)
        {
            return (int?)GetInternalRecursive(obj, expr.Body);
        }
        public static bool? Get<T>(this T obj, Expression<Func<T, bool>> expr)
        {
            return (bool?)GetInternalRecursive(obj, expr.Body);
        }
        public static TV Get<T, TV>(this T obj, Expression<Func<T, TV>> expr)
        {
            return (TV)GetInternalRecursive(obj, expr.Body);
        }
        public static TV? Get<T, TV>(this T obj, Expression<Func<T, TV?>> expr)
            where TV : struct
        {
            return (TV?)GetInternalRecursive(obj, expr.Body);
        }
    }
}
